/*
 * 29/03/2004
 *
 * math_utils.h
 * Fonctions et constantes mathmatiques utiles
 *
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

#ifndef _MATH_UTILS_H_
#define _MATH_UTILS_H_

#include <math.h>

#define PI 3.14159265358979323846f

// This code supposedly originates from Id-software
// It makes a fast 1 / sqrtf(v) approximation
inline float rsqrtf(float v){
    float v_half = v * 0.5f;
    long i = *(long *) &v;
    i = 0x5f3759df - (i >> 1);
    v = *(float *) &i;
    return v * (1.5f - v_half * v * v);
}

inline float sqrf(const float x){
    return x * x;
}

inline float sincf(const float x){
	return (x == 0)? 1 : sinf(x) / x;
}

#define roundf(x) floorf((x) + 0.5f)

inline float intAdjustf(const float x, const float diff = 0.01f){
	float f = roundf(x);

	return (fabsf(f - x) < diff)? f : x;
}

inline unsigned int getClosestPowerOfTwo(const unsigned int x){
	int i,k;
	
	k = x;
	i = -1;
	while (k != 0){
		k >>= 1;
		i++;
	}
	return 1 << (i + ((x >> (i-1)) & 1));
}

inline unsigned int getNextPowerOfTwo(const unsigned int x) {

	if (x<= 2) return 2;
	if (x<= 16) return 16;
	if (x<= 32) return 32;
	if (x<= 64) return 64;
	if (x<= 128) return 128;
	if (x<= 256) return 256;
	if (x<= 512) return 512;
	if (x<= 1024) return 1024;
	return 0;
}

// Inversion d'une matrice 4*4
inline void matrixInvert(float mat[16])
{
	float temp;

	temp=mat[1];
	mat[1]=mat[4];
	mat[4]=temp;

	temp=mat[2];
	mat[2]=mat[8];
	mat[8]=temp;

	temp=mat[6];
	mat[6]=mat[9];
	mat[9]=temp;
  
	mat[12]=-mat[12];
	mat[13]=-mat[13];
	mat[14]=-mat[14];
}

// prints the contents of a 16 element array to stdout
inline void matrixDebug (GLfloat matrix[16])
{
   fprintf(stdout, "|%f %f %f %f|\n", matrix[0], matrix[1], matrix[2], matrix[3]);
   fprintf(stdout, "|%f %f %f %f|\n", matrix[4], matrix[5], matrix[6], matrix[7]);
   fprintf(stdout, "|%f %f %f %f|\n", matrix[8], matrix[9], matrix[10], matrix[11]);
   fprintf(stdout, "|%f %f %f %f|\n\n", matrix[12], matrix[13], matrix[14], matrix[15]);

   return;
}

// prints the contents of a 4 element array to stdout
inline void vectorDebug (GLfloat vector[4])
{
   fprintf(stdout, "<%f, %f, %f, %f>\n", vector[0], vector[1], vector[2], vector[3]);
   
   return;
}

//////////////////////////
// Set matrix to identity
//   matrix - Input/Output matrix
//////////////////////////
inline void 
matrixIdentity(GLfloat matrix[16])
{
  matrix[ 0] = 1.0;  matrix[ 1] = 0.0;  matrix[ 2] = 0.0;  matrix[ 3] = 0.0;
  matrix[ 4] = 0.0;  matrix[ 5] = 1.0;  matrix[ 6] = 0.0;  matrix[ 7] = 0.0;
  matrix[ 8] = 0.0;  matrix[ 9] = 0.0;  matrix[10] = 1.0;  matrix[11] = 0.0;
  matrix[12] = 0.0;  matrix[13] = 0.0;  matrix[14] = 0.0;  matrix[15] = 1.0;
}

//////////////////////////
// Invert a matrix. (Matrix MUST be orhtonormal!)
//   in - Input matrix
//   out - Output matrix
//////////////////////////
inline void
matrixInvert(GLfloat in[16], GLfloat out[16])
{
  // Transpose rotation
  out[ 0] = in[ 0];  out[ 1] = in[ 4];  out[ 2] = in[ 8];
  out[ 4] = in[ 1];  out[ 5] = in[ 5];  out[ 6] = in[ 9];
  out[ 8] = in[ 2];  out[ 9] = in[ 6];  out[10] = in[10];
  
  // Clear shearing terms
  out[3] = 0.0f; out[7] = 0.0f; out[11] = 0.0f; out[15] = 1.0f;

  // Translation is minus the dot of tranlation and rotations
  out[12] = -(in[12]*in[ 0]) - (in[13]*in[ 1]) - (in[14]*in[ 2]);
  out[13] = -(in[12]*in[ 4]) - (in[13]*in[ 5]) - (in[14]*in[ 6]);
  out[14] = -(in[12]*in[ 8]) - (in[13]*in[ 9]) - (in[14]*in[10]);
}

//////////////////////////
// Multiply a vector by a matrix.
//   vecIn - Input vector
//   m - Input matrix
//   vecOut - Output vector
//////////////////////////
inline void
vecMatMult(GLfloat vecIn[3], GLfloat m[16], GLfloat vecOut[3]) {
  vecOut[0] = (vecIn[0]*m[ 0]) + (vecIn[1]*m[ 4]) + (vecIn[2]*m[ 8]) + m[12];
  vecOut[1] = (vecIn[0]*m[ 1]) + (vecIn[1]*m[ 5]) + (vecIn[2]*m[ 9]) + m[13];
  vecOut[2] = (vecIn[0]*m[ 2]) + (vecIn[1]*m[ 6]) + (vecIn[2]*m[10]) + m[14];
}


//////////////////////////
// Multiply a vector by just the 3x3 portion of a matrix.
//   vecIn - Input vector
//   m - Input matrix
//   vecOut - Output vector
//////////////////////////
inline void
vecMat3x3Mult(GLfloat vecIn[3], GLfloat m[16], GLfloat vecOut[3]) {
  vecOut[0] = (vecIn[0]*m[ 0]) + (vecIn[1]*m[ 4]) + (vecIn[2]*m[ 8]);
  vecOut[1] = (vecIn[0]*m[ 1]) + (vecIn[1]*m[ 5]) + (vecIn[2]*m[ 9]);
  vecOut[2] = (vecIn[0]*m[ 2]) + (vecIn[1]*m[ 6]) + (vecIn[2]*m[10]);
}



//
// 
// project x,y onto a hemi-sphere centered within width, height
//
////////////////////////////////////////////////////////////
inline void ptov(int x, int y, int width, int height, float v[3]){
  float d, a;

  // project x,y onto a hemi-sphere centered within width, height
  v[0] = (2.0f*x - width) / width;
  v[1] = (height - 2.0f*y) / height;
  d = sqrt(v[0]*v[0] + v[1]*v[1]);
  v[2] = cos((PI/2.0f) * ((d < 1.0f) ? d : 1.0f));
  a = 1.0f / sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  v[0] *= a;
  v[1] *= a;
  v[2] *= a;
}



#endif // _MYMATH_H_
